{** Unit FloatCtrl.
 *  @author (c) 2001-2003 fixxion freeware, Hamburg
 *}
//==============================================================================
unit FloatCtrl;
//==============================================================================

//==============================================================================
interface
//==============================================================================

//------------------------------------------------------------------------------
{** @description Asymmetric rounding method, which rounds always to the
 *  next bigger number, if a value is just in the middle of two numbers.<br>
 *  <br>
 *  Notes:<br>
 *   Round control settings of the float processing unit (FPU) are ignored.<br>
 *   Results are limited by hardware and software precision capabilities and
 *    settings.<br>
 *   Use standard Round function for symmetric rounding to the next even
 *    number.<br>
 *  <br>
 *  Examples:<br>
 *  <code>
 *  -2.5 rounds to -2<br>
 *  -1.5 rounds to -1<br>
 *  -0.5 rounds to -0<br>
 *   0.5 rounds to  1<br>
 *   1.5 rounds to  2<br>
 *   2.5 rounds to  3<br>
 *  </code>
 *}
function RoundJustified(aValue: Extended): integer;

//------------------------------------------------------------------------------
const
  FL_EXC_INVALID_OPERATION    = $0;
  FL_EXC_DENORMALIZED_OPERAND = $1;
  FL_EXC_ZERO_DIVIDE          = $2;
  FL_EXC_OVERFLOW             = $3;
  FL_EXC_UNDERFLOW            = $4;
  FL_EXC_PRECISION            = $5;
  FL_EXC_INTERRUPT            = $7;
  FL_PRECISION_CONTROL_LOW    = $8;
  FL_PRECISION_CONTROL_HIGH   = $9;
  FL_ROUND_CONTROL_LOW        = $A;
  FL_ROUND_CONTROL_HIGH       = $B;
  FL_INFINITY_CONTROL         = $C;

//------------------------------------------------------------------------------
{** @summary The ControlWord property can be used for access to the control word
 *  register of the float processing unit (FPU).
 *  @description The original register value is stored when the class in its
 *  Create and restored in Destroy method.
 *}
type
  TFloatControl = class
  private
    fControlWord: word;
    fControlWordSaved: word;
    function  GetControlWord: word;
    procedure SetControlWord(aValue: word);
  public
    constructor Create;
    destructor  Destroy; override;
  published
    property ControlWord: word
      read     GetControlWord
      write    SetControlWord;
  end;

//------------------------------------------------------------------------------
{** @summary The properties can be used for detailed access the control word
 *  register of the float processing unit (FPU).
 *}
type
  TFloatControlDetailed = class(TFloatControl)
  private
    function GetExcInvalidOperation: boolean;
    procedure SetExcInvalidOperation(aValue: boolean);
    function GetExcDenormalizedOperand: boolean;
    procedure SetExcDenormalizedOperand(aValue: boolean);
    function GetExcZeroDivide: boolean;
    procedure SetExcZeroDivide(aValue: boolean);
    function GetExcOverflow: boolean;
    procedure SetExcOverflow(aValue: boolean);
    function GetExcUnderflow: boolean;
    procedure SetExcUnderflow(aValue: boolean);
    function GetExcPrecision: boolean;
    procedure SetExcPrecision(aValue: boolean);
    function GetExcInterrupt: boolean;
    procedure SetExcInterrupt(aValue: boolean);
    function GetPrecisionControlLow: boolean;
    procedure SetPrecisionControlLow(aValue: boolean);
    function GetPrecisionControlHigh: boolean;
    procedure SetPrecisionControlHigh(aValue: boolean);
    function GetRoundControlLow: boolean;
    procedure SetRoundControlLow(aValue: boolean);
    function GetRoundControlHigh: boolean;
    procedure SetRoundControlHigh(aValue: boolean);
    function GetInfinityControl: boolean;
    procedure SetInfinityControl(aValue: boolean);
  public
    function ControlWordIsBitSet(aBitIndex: integer): boolean;
    procedure ControlWordSetBit(
      aBitIndex: integer;
      const aSetBit: boolean = TRUE);
  published
    property ExcInvalidOperation: boolean
      read     GetExcInvalidOperation
      write    SetExcInvalidOperation;
    property ExcDenormalizedOperand: boolean
      read     GetExcDenormalizedOperand
      write    SetExcDenormalizedOperand;
    property ExcZeroDivide: boolean
      read     GetExcZeroDivide
      write    SetExcZeroDivide;
    property ExcOverflow: boolean
      read     GetExcOverflow
      write    SetExcOverflow;
    property ExcUnderflow: boolean
      read     GetExcUnderflow
      write    SetExcUnderflow;
    property ExcPrecision: boolean
      read     GetExcPrecision
      write    SetExcPrecision;
    property ExcInterrupt: boolean
      read     GetExcInterrupt
      write    SetExcInterrupt;
    property PrecisionControlLow: boolean
      read     GetPrecisionControlLow
      write    SetPrecisionControlLow;
    property PrecisionControlHigh: boolean
      read     GetPrecisionControlHigh
      write    SetPrecisionControlHigh;
    property RoundControlLow: boolean
      read     GetRoundControlLow
      write    SetRoundControlLow;
    property RoundControlHigh: boolean
      read     GetRoundControlHigh
      write    SetRoundControlHigh;
    property InfinityControl: boolean
      read     GetInfinityControl
      write    SetInfinityControl;
  end;

//------------------------------------------------------------------------------
type
  TFloatControlStrings =
    array[FL_EXC_INVALID_OPERATION..FL_INFINITY_CONTROL] of string;
//------------------------------------------------------------------------------
const
  FLOAT_CONTROL_NAMES: TFloatControlStrings = (
    'Invalid operation',
    'Denormalized operand',
    'Zero divide',
    'Overflow',
    'Underflow',
    'Precision',
    '(none)',
    'Exception interrupt',
    'Precision control low',
    'Precision control high',
    'Round control low',
    'Round control high',
    'Infinity control'
  );
//------------------------------------------------------------------------------
const
  FLOAT_CONTROL_HINTS:
    TFloatControlStrings = (
    'Attempt to use an unknown operation',
    'Attempt to operate with a denormalized value',
    'Attempt to divide by zero',
    'The result of the previous operation is too large',
    'The result of the previous operation is too small',
    'The precision of the previous operation was to small',
    '(none)',
    'Interrupt on Exception',
    'Precision control low bit',
    'Precision control high bit',
    'Round control low bit',
    'Round control high bit',
    'Infinity control'
  );

//------------------------------------------------------------------------------
type
  TFourStrings =
    array[0..3] of string;
const
  FLOAT_PRECISION_CONTROL_NAMES: TFourStrings = (
    'Single (24 bit)',
    'Reserved (not used)',
    'Double (53 bit)',
    'Extended (64 bit)'
  );
const
  FLOAT_ROUND_CONTROL_NAMES: TFourStrings = (
    'Next or even value',
    'Negative infinity',
    'Positive infinity',
    'Truncate'
  );

//------------------------------------------------------------------------------
type
  TBoolStrings =
    array[boolean] of string;
const
  FLOAT_INFINITY_CONTROL_NAMES: TBoolStrings = (
    'Projective model',
    'Affine model'
  );
  FLOAT_INFINITY_CONTROL_HINTS: TBoolStrings = (
    'Numbers in a circle with -INF = +INF',
    'Numbers in a line from -INF .. +INF'
  );

//------------------------------------------------------------------------------
{** @summary The functions can be used to get verbose information about some
 *  control word register flags.
 *}
type
  TFloatControlVerbose = class(TFloatControlDetailed)
  public
    function PrecisionControlStr: string;
    function RoundControlStr: string;
    function InfinityControlStr: string;
    function InfinityControlHint: string;
  end;

//==============================================================================
implementation
//==============================================================================

//------------------------------------------------------------------------------
function RoundJustified(aValue: Extended): integer;
begin
  Result := Trunc(aValue) + Trunc( 2 * Frac(aValue));
  if ( aValue < 0 ) then // Negative values...
    if ( (Frac(aValue) * 1000000000) = -500000000 ) then // Justify...
      Inc(Result);
end; // RoundJustified

//------------------------------------------------------------------------------
// TFloatControl
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
constructor TFloatControl.Create;
begin
  inherited;
  fControlWordSaved := Default8087CW;
  fControlWord := Default8087CW;
end;

//------------------------------------------------------------------------------
destructor  TFloatControl.Destroy;
begin
  Set8087CW(fControlWordSaved);
  inherited;
end;

//------------------------------------------------------------------------------
function  TFloatControl.GetControlWord: word;
begin
  Result := fControlWord;
end;

//------------------------------------------------------------------------------
procedure TFloatControl.SetControlWord(aValue: word);
begin
  if ( fControlWord <> aValue ) then
  begin
    fControlWord := aValue;
    Set8087CW(fControlWord);
  end;
end;

//------------------------------------------------------------------------------
// TFloatControlDetailed
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetExcInvalidOperation: boolean;
begin
  Result := (ControlWord and (1 shl FL_EXC_INVALID_OPERATION) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetExcInvalidOperation(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_EXC_INVALID_OPERATION)
  else
    ControlWord := ControlWord and not (1 shl FL_EXC_INVALID_OPERATION);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetExcDenormalizedOperand: boolean;
begin
  Result := (ControlWord and (1 shl FL_EXC_DENORMALIZED_OPERAND) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetExcDenormalizedOperand(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_EXC_DENORMALIZED_OPERAND)
  else
    ControlWord := ControlWord and not (1 shl FL_EXC_DENORMALIZED_OPERAND);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetExcZeroDivide: boolean;
begin
  Result := (ControlWord and (1 shl FL_EXC_ZERO_DIVIDE) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetExcZeroDivide(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_EXC_ZERO_DIVIDE)
  else
    ControlWord := ControlWord and not (1 shl FL_EXC_ZERO_DIVIDE);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetExcOverflow: boolean;
begin
  Result := (ControlWord and (1 shl FL_EXC_OVERFLOW) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetExcOverflow(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_EXC_OVERFLOW)
  else
    ControlWord := ControlWord and not (1 shl FL_EXC_OVERFLOW);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetExcUnderflow: boolean;
begin
  Result := (ControlWord and (1 shl FL_EXC_UNDERFLOW) <> 0);
end;

procedure TFloatControlDetailed.SetExcUnderflow(aValue: boolean);
//------------------------------------------------------------------------------
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_EXC_UNDERFLOW)
  else
    ControlWord := ControlWord and not (1 shl FL_EXC_UNDERFLOW);
end;

function TFloatControlDetailed.GetExcPrecision: boolean;
//------------------------------------------------------------------------------
begin
  Result := (ControlWord and (1 shl FL_EXC_PRECISION) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetExcPrecision(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_EXC_PRECISION)
  else
    ControlWord := ControlWord and not (1 shl FL_EXC_PRECISION);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetExcInterrupt: boolean;
begin
  Result := (ControlWord and (1 shl FL_EXC_INTERRUPT) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetExcInterrupt(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_EXC_INTERRUPT)
  else
    ControlWord := ControlWord and not (1 shl FL_EXC_INTERRUPT);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetPrecisionControlLow: boolean;
begin
  Result := (ControlWord and (1 shl FL_PRECISION_CONTROL_LOW) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetPrecisionControlLow(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_PRECISION_CONTROL_LOW)
  else
    ControlWord := ControlWord and not (1 shl FL_PRECISION_CONTROL_LOW);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetPrecisionControlHigh: boolean;
begin
  Result := (ControlWord and (1 shl FL_PRECISION_CONTROL_HIGH) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetPrecisionControlHigh(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_PRECISION_CONTROL_HIGH)
  else
    ControlWord := ControlWord and not (1 shl FL_PRECISION_CONTROL_HIGH);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetRoundControlLow: boolean;
begin
  Result := (ControlWord and (1 shl FL_ROUND_CONTROL_LOW) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetRoundControlLow(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_ROUND_CONTROL_LOW)
  else
    ControlWord := ControlWord and not (1 shl FL_ROUND_CONTROL_LOW);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetRoundControlHigh: boolean;
begin
  Result := (ControlWord and (1 shl FL_ROUND_CONTROL_HIGH) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetRoundControlHigh(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_ROUND_CONTROL_HIGH)
  else
    ControlWord := ControlWord and not (1 shl FL_ROUND_CONTROL_HIGH);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.GetInfinityControl: boolean;
begin
  Result := (ControlWord and (1 shl FL_INFINITY_CONTROL) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.SetInfinityControl(aValue: boolean);
begin
  if aValue then
    ControlWord := ControlWord or (1 shl FL_INFINITY_CONTROL)
  else
    ControlWord := ControlWord and not (1 shl FL_INFINITY_CONTROL);
end;

//------------------------------------------------------------------------------
function TFloatControlDetailed.ControlWordIsBitSet(
  aBitIndex: integer): boolean;
begin
  Result := FALSE;
  if aBitIndex in [FL_EXC_INVALID_OPERATION .. FL_INFINITY_CONTROL] then
    Result := (ControlWord and (1 shl aBitIndex) <> 0);
end;

//------------------------------------------------------------------------------
procedure TFloatControlDetailed.ControlWordSetBit(
  aBitIndex: integer;
  const aSetBit: boolean = TRUE);
begin
  if aBitIndex in [FL_EXC_INVALID_OPERATION .. FL_INFINITY_CONTROL] then
    if aSetBit then
      ControlWord := ControlWord or (1 shl aBitIndex)
    else
      ControlWord := ControlWord and not (1 shl aBitIndex);
end;

//------------------------------------------------------------------------------
// TFloatControlVerbose
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
function TFloatControlVerbose.PrecisionControlStr: string;
var
  a_Value: integer;
begin
  a_Value := 0;
  if PrecisionControlLow then
    a_Value := 1;
  if PrecisionControlHigh then
    a_Value := a_Value + 1 shl 1;
  Result := FLOAT_PRECISION_CONTROL_NAMES[a_Value];
end;

//------------------------------------------------------------------------------
function TFloatControlVerbose.RoundControlStr: string;
var
  a_Value: integer;
begin
  a_Value := 0;
  if RoundControlLow then
    a_Value := 1;
  if RoundControlHigh then
    a_Value := a_Value + 1 shl 1;
  Result := FLOAT_ROUND_CONTROL_NAMES[a_Value];
end;

//------------------------------------------------------------------------------
function TFloatControlVerbose.InfinityControlStr: string;
begin
  Result := FLOAT_INFINITY_CONTROL_NAMES[InfinityControl];
end;

//------------------------------------------------------------------------------
function TFloatControlVerbose.InfinityControlHint: string;
begin
  Result := FLOAT_INFINITY_CONTROL_HINTS[InfinityControl];
end;

end.
